/**
* octomap_server: A ROS map server for the 3D Octomaps
* (inspired by the ROS map_server)
* @author A. Hornung, University of Freiburg, Copyright (C) 2009.
* @see http://octomap.sourceforge.net/
* License: GNU GPL v2, http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt
*/

/*
 * Copyright (c) 2010, A. Hornung, University of Freiburg
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of Freiburg nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <ros/ros.h>
#include <octomap_ros/GetOctomap.h>
#include <octomap_ros/conversions.h>
#include <octomap/octomap.h>
#include <fstream>

#define USAGE "\nUSAGE: octomap_saver <map.bt>\n" \
              "  map.bt: filename of map to be saved\n"

using namespace std;
 
/**
 * @brief Map generation node.
 */
class MapSaver{
  public:
    MapSaver(const std::string& mapname){
      ros::NodeHandle n;
      const static std::string servname = "octomap_binary";
      ROS_INFO("Requesting the map from %s...", n.resolveName(servname).c_str());
      octomap_ros::GetOctomap::Request req;
      octomap_ros::GetOctomap::Response resp;
      while(n.ok() && !ros::service::call(servname, req, resp))
      {
        ROS_WARN("Request to %s failed; trying again...", n.resolveName(servname).c_str());
        usleep(1000000);
      }

      if (n.ok()){ // skip when CTRL-C
		  ROS_INFO("Map received, saving to %s", mapname.c_str());
		  ofstream mapfile(mapname.c_str(), ios_base::binary);

		  if (!mapfile.is_open()){
			  ROS_ERROR("Could not open file %s for writing", mapname.c_str());
		  } else {
			  // test conversion:
//			  octomap::OcTree octomap(0.1);
//			  octomap_server::octomapMsgToMap(resp.map, octomap);
//			  octomap.writeBinary(mapname);

			  // write out stream directly
			  mapfile.write((char*)&resp.map.data[0], resp.map.data.size());
			  mapfile.close();
		  }
      }
  }
};

int main(int argc, char** argv){
  ros::init(argc, argv, "octomap_saver");
  std::string mapFilename("");
  if (argc == 2)
	  mapFilename = std::string(argv[1]);
  else{
	  ROS_ERROR("%s", USAGE);
	  exit(-1);
  }

  try{
	  MapSaver ms(mapFilename);
  }catch(std::runtime_error& e){
	  ROS_ERROR("map_saver exception: %s", e.what());
	  return -1;
  }

  return 0;
}


